
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <initng.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include <initng_global.h>
#include <initng_process_db.h>
#include <initng_service_cache.h>
#include <initng_handler.h>
#include <main.h>
#include <initng_active_db.h>
#include <initng_toolbox.h>
#include <initng_plugin_hook.h>
#include <initng_load_module.h>
#include <initng_plugin_callers.h>
#include <initng_global.h>
#include <initng_error.h>
#include <initng_string_tools.h>
#include <initng_static_states.h>
#include <initng_depend.h>
#include <initng_main.h>

#include "print_service.h"


static void cmd_fast_reload(char *arg);
static int cmd_get_pid_of(char *arg);
static int cmd_start_on_new(char *arg);
static int cmd_free_service(char *arg);
static int cmd_restart(char *arg);
static char *cmd_print_uptime(char *arg);
static int cmd_reload(char *arg);
static int cmd_initng_quit(char *arg);
static int cmd_initng_halt(char *arg);
static int cmd_initng_reboot(char *arg);
static char *cmd_print_modules(char *arg);
static int cmd_load_module(char *arg);
static int cmd_unload_module(char *arg);
static char *cmd_print_service_db(char *arg);
static char *cmd_print_active_db(char *arg);
static int cmd_percent_done(char *arg);
static char *cmd_get_father_of(char *arg);
static char *cmd_get_depends_on(char *arg);
static char *cmd_get_depends_on_deep(char *arg);
static char *cmd_get_depends_off(char *arg);
static char *cmd_get_depends_off_deep(char *arg);
static int cmd_new_init(char *arg);
static int cmd_runlevel(char *arg);
static char *cmd_print_fds(char *arg);

#ifdef DEBUG
static int cmd_toggle_verbose(char *arg);
static int cmd_add_verbose(char *arg);
static int cmd_del_verbose(char *arg);
#endif

s_command LIST_FDS = { 'I', "list_filedescriptors", STRING_COMMAND, ADVANCHED_COMMAND, NO_OPT, {(void *) &cmd_print_fds}, "Print all open filedescriptors initng have." };

s_command FAST_RELOAD = { 'c', "hot_reload", VOID_COMMAND, STANDARD_COMMAND, NO_OPT,
    {(void *) &cmd_fast_reload},
    "Fast Reload"
};
s_command GET_PID_OF = { 'g', "get_pid_of", INT_COMMAND, ADVANCHED_COMMAND, REQUIRES_OPT,
    {(void *) &cmd_get_pid_of}, "Get pid of service"
};
s_command START_ON_NEW = { 'j', "restart_from", TRUE_OR_FALSE_COMMAND, ADVANCHED_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_start_on_new},
    "Stop all services, and start from"
};
s_command FREE_SERVICE = { 'z', "zap", TRUE_OR_FALSE_COMMAND, STANDARD_COMMAND, USES_OPT,
    {(void *) &cmd_free_service},
    "Resets a failed service, so it can be restarted."
};
s_command RESTART_SERVICE = { 'r', "restart", TRUE_OR_FALSE_COMMAND, STANDARD_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_restart},
    "Restart service"
};

s_command CHANGE_RUNLEVEL = { 'e', "runlevel", TRUE_OR_FALSE_COMMAND, STANDARD_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_runlevel},
    "Change Runlevel"
};

s_command PRINT_UPTIME = { 't', "time", STRING_COMMAND, ADVANCHED_COMMAND, REQUIRES_OPT,
    {(void *) &cmd_print_uptime},
    "Print uptime"
};
s_command SERVICE_RELOAD = { 'R', "reload_service", TRUE_OR_FALSE_COMMAND, STANDARD_COMMAND,
    USES_OPT,
    {(void *) &cmd_reload},
    "Reload service data from disk ( reparse /etc/initng )"
};

s_command QUIT_INITNG = { 'q', "quit", TRUE_OR_FALSE_COMMAND, ADVANCHED_COMMAND, NO_OPT,
    {(void *) &cmd_initng_quit},
    "Quits initng"
};
s_command HALT_INITNG = { '0', "halt", TRUE_OR_FALSE_COMMAND, STANDARD_COMMAND, NO_OPT,
    {(void *) &cmd_initng_halt},
    "Halt the computer"
};
s_command REBOOT_INITNG = { '6', "reboot", TRUE_OR_FALSE_COMMAND, STANDARD_COMMAND, NO_OPT,
    {(void *) &cmd_initng_reboot},
    "Reboot the computer"
};


s_command PRINT_MODULES = { 'm', "print_plugins", STRING_COMMAND, ADVANCHED_COMMAND, NO_OPT,
    {(void *) &cmd_print_modules},
    "Print loaded plugins"
};

s_command LOAD_MODULE = { 'o', "load_module", TRUE_OR_FALSE_COMMAND, ADVANCHED_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_load_module},
    "Load Module"
};
s_command UNLOAD_MODULE = { 'w', "unload_module", TRUE_OR_FALSE_COMMAND, ADVANCHED_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_unload_module},
    "UnLoad Module"
};
s_command PRINT_SERVICE_DB = { 'p', "print_service_db", STRING_COMMAND, ADVANCHED_COMMAND,
    USES_OPT,
    {(void *) &cmd_print_service_db},
    "Print service_db"
};

s_command PRINT_ACTIVE_DB = { 'P', "print_active_db", STRING_COMMAND, ADVANCHED_COMMAND, USES_OPT,
    {(void *) &cmd_print_active_db},
    "Print active_db"
};

s_command PERCENT_DONE = { 'n', "done", INT_COMMAND, ADVANCHED_COMMAND, NO_OPT,
    {(void *) &cmd_percent_done},
    "Prints percent of system up"
};

s_command GET_FATHER_OF = { 'f', "father", STRING_COMMAND, ADVANCHED_COMMAND, REQUIRES_OPT,
    {(void *) &cmd_get_father_of},
    "Print father to"
};

s_command DEPENDS_ON = { 'a', "service_dep_on", STRING_COMMAND, ADVANCHED_COMMAND, REQUIRES_OPT,
    {(void *) &cmd_get_depends_on},
    "Print what services me depends on"
};

s_command DEPENDS_ON_DEEP = { 'A', "service_dep_on_deep", STRING_COMMAND, ADVANCHED_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_get_depends_on_deep},
    "Print what services me depends on deep"
};

s_command DEPENDS_OFF = { 'b', "service_dep_on_me", STRING_COMMAND, ADVANCHED_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_get_depends_off},
    "Print what dependencies that are depending on me"
};

s_command DEPENDS_OFF_DEEP = { 'B', "service_dep_on_me_deep", STRING_COMMAND, ADVANCHED_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_get_depends_off_deep},
    "Print what dependencies that are depending on me deep"
};


s_command NEW_INIT = { 'E', "new_init", TRUE_OR_FALSE_COMMAND, HIDDEN_COMMAND, REQUIRES_OPT,
    {(void *) &cmd_new_init},
    "Stops all services, and when its done, launching a new init."
};

#ifdef DEBUG
s_command TOGGLE_VERBOSE = { 'v', "verbose", TRUE_OR_FALSE_COMMAND, ADVANCHED_COMMAND, NO_OPT,
    {(void *) &cmd_toggle_verbose},
    "Toggle the verbose flag - ONLY FOR DEBUGGING"
};
s_command ADD_VERBOSE = { 'i', "add_verbose", TRUE_OR_FALSE_COMMAND, ADVANCHED_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_add_verbose},
    "Add string to watch for to make initng verbose - ONLY FOR DEBUGGING"
};
s_command DEL_VERBOSE = { 'k', "del_verbose", TRUE_OR_FALSE_COMMAND, ADVANCHED_COMMAND,
    REQUIRES_OPT,
    {(void *) &cmd_del_verbose},
    "Del string to watch for to make initng verbose - ONLY FOR DEBUGGING"
};
#endif


static int cmd_get_pid_of(char *arg)
{
    active_db_h *apt = NULL;
    process_h *process = NULL;

    if (!arg)
        return (-2);


    if (!(apt = initng_active_db_find_in_name(arg)))
        return (-1);

    /* browse all processes */
    while_processes(process, apt)
    {
        /* return the first process found */
        return (process->pid);
    }

    return (-3);
}


static void cmd_fast_reload(char *arg)
{
    (void) arg;
    int retval;
    char *new_argv[4];

    retval = initng_plugin_callers_dump_state();

    if (retval == TRUE)
    {
        D_("exec()int initng\n");
        new_argv[0] = i_strdup("/sbin/initng");
        new_argv[1] = i_strdup("--hot_reload");
        new_argv[2] = NULL;

        execve("/sbin/initng", new_argv, environ);
        F_("Failed to reload initng!\n");
    }
    else
    {
        if (retval == FALSE)
            F_("No plugin was willing to dump state\n");
        else
            F_("dump_state failed!\n");
    }
}

static int cmd_start_on_new(char *arg)
{
    if (!arg)
        return (FALSE);

    g.when_out = THEN_RESTART;
    initng_main_set_runlevel(arg);
    initng_handler_stop_all();
    return (TRUE);
}

static int cmd_free_service(char *arg)
{
    active_db_h *apt = NULL;

    if (!arg)
        return (FALSE);

    if (!(apt = initng_active_db_find_in_name(arg)))
        return (FALSE);

    initng_active_db_del(apt);
    initng_active_db_free(apt);

    /* also flush file cache */
    cmd_reload(arg);

    /* return happily */
    return (TRUE);
}


static int cmd_restart(char *arg)
{
    active_db_h *apt = NULL;

    if (!arg)
    {
        F_("Must tell the service name to restart.\n");
        return (FALSE);
    }

    apt = initng_active_db_find_in_name(arg);
    if (!apt)
    {
        return (FALSE);
        F_("Service \"%s\" not found.\n", arg);
    }

    D_("removing service data for %s, to make sure .ii file is reloaded!\n",
       arg);
    if (apt->from_service)
    {
        list_del(&apt->from_service->list);
        initng_service_cache_free(apt->from_service);
    }

    D_("Restarting service %s\n", apt->name);
    return (initng_handler_restart_service(apt));
}





static char *cmd_print_uptime(char *arg)
{
    active_db_h *apt = NULL;
    struct timeval now;
    char *string;

    if (!arg)
    {
        return(i_strdup("Please tell me what service to get up-time from."));
    }

    apt = initng_active_db_find_in_name(arg);
    if (!apt)
    {
	string=i_calloc(35+strlen(arg), sizeof(char));
        sprintf(string, "Service \"%s\" is not found!", arg);
        return(string);
    }

    gettimeofday(&now, NULL);
    {
	string=i_calloc(50, sizeof(char));
	sprintf(string, "Up-time is %ims.\n", MS_DIFF(now, apt->time_current_state));
	return(string);
    }
}


static int cmd_reload(char *arg)
{
    service_cache_h *s;

    /* if no arg, or empty arg set, remove all content */
    if (!arg || strlen(arg) < 2)
    {
        D_("Clearing complete db.\n");
        initng_service_cache_free_all();
        return (TRUE);
    }

    s = initng_service_cache_find_in_name(arg);
    if (!s)
    {
        D_("Did not find service %s to release cache for!\n", arg);
        return (FALSE);
    }

    D_("removing service data for %s!\n", arg);
    list_del(&s->list);
    initng_service_cache_free(s);
    return (TRUE);
}

static int cmd_initng_reboot(char *arg)
{
    (void) arg;
    g.when_out = THEN_REBOOT;
    initng_handler_stop_all();
    return (TRUE);
}

static int cmd_initng_halt(char *arg)
{
    (void) arg;
    g.when_out = THEN_HALT;
    initng_handler_stop_all();
    return (TRUE);
}

static int cmd_initng_quit(char *arg)
{
    (void) arg;
    g.when_out = THEN_QUIT;
    initng_handler_stop_all();
    return (TRUE);
}



static char *cmd_print_modules(char *arg)
{
    m_h *mod = NULL;
    size_t string_len=20;
    char *string=i_calloc(string_len, sizeof(char));

    (void) arg;


    sprintf(string, "modules: \n");

    while_module_db(mod)
    {
	/* Increase buffer for adding */
	string_len+=strlen(mod->module_name) + 8;
	string=i_realloc(string, string_len * sizeof(char));
	strcat(string, "  * ");
	strcat(string, mod->module_name);
	strcat(string, "\n");
    }

    return(string);
}

static int cmd_load_module(char *arg)
{
    if (!arg)
        return (FALSE);

    /* load the module */
    if (initng_load_module(arg) == NULL)
        return (FALSE);

    /* clear the service cache, so new variables will be read next time */
    initng_service_cache_free_all();

    return (TRUE);
}

static int cmd_unload_module(char *arg)
{
    if (!arg)
        return (FALSE);

    /* Just to make sure. */
    initng_service_cache_free_all();

    return (initng_unload_module_named(arg));
}

static char *cmd_print_service_db(char *arg)
{
    service_cache_h *s;

    D_("Print service \"%s\"\n", arg);
    if (arg && strlen(arg) > 1
        && (s = initng_service_cache_find_in_name(arg)))
    {
        return(service_db_print(s));
    }
    else if (arg && strlen(arg) > 1)
    {
	return(i_strdup("No such service.\n"));
    }
    else
    {
        return(service_db_print_all());
    }

    return(NULL);
}

static char *cmd_print_active_db(char *arg)
{
    active_db_h *s;

    if (arg && strlen(arg) > 1 && (s = initng_active_db_find_in_name(arg)))
    {
        return(active_db_print(s));
    }
    else if (arg && strlen(arg) > 1)
    {
	return(i_strdup("No such service."));
    }
    else
    {
        return(active_db_print_all());
    }

    return(NULL);
}

static int cmd_percent_done(char *arg)
{
    (void) arg;

    return (initng_active_db_percent_started());
}



static char *cmd_get_father_of(char *arg)
{
    char *string = NULL;
    service_cache_h *s;

    s = initng_service_cache_find_in_name(arg);
    if (!s)
    {
	return(i_strdup("Can't find service."));
    }

    if (s->father_name)
        mprintf(&string, "Father to service %s is %s\n", s->name, s->father_name);
    else
        mprintf(&string, "Service %s has no father\n", s->name);

    return(string);
}


static char *cmd_get_depends_on(char *arg)
{
    char *string = NULL;
    active_db_h *current = NULL;
    active_db_h *on = NULL;

    on = initng_active_db_find_in_name(arg);

    if (!on)
	return(i_strdup("Did not find service."));

    mprintf(&string, "The \"%s\" depends on:\n", on->name);

    while_active_db(current)
    {
        /* if that we are watching needs current */
        if (initng_depend(on, current) == TRUE)
        {
            if (current->current_state && current->current_state->state_name)
            {
                mprintf(&string, "  %s\t\t[%s]\n", current->name,
                        current->current_state->state_name);
            }
        }
    }
    return(string);
}

static char *cmd_get_depends_on_deep(char *arg)
{
    char *string = NULL;
    active_db_h *current = NULL;
    active_db_h *on = NULL;

    on = initng_active_db_find_in_name(arg);

    if (!on)
	return(i_strdup("Did not find service."));

    mprintf(&string, "The \"%s\" depends on:\n", on->name);

    while_active_db(current)
    {
        /* if that we are watching needs current */
        if (initng_depend_deep(on, current) == TRUE)
        {
            if (current->current_state && current->current_state->state_name)
            {
                mprintf(&string, "  %s\t\t[%s]\n", current->name,
                        current->current_state->state_name);
            }
        }
    }
    return(string);
}

static char *cmd_get_depends_off(char *arg)
{
    char *string = NULL;
    active_db_h *current = NULL;
    active_db_h *on = NULL;

    on = initng_active_db_find_in_name(arg);
    if (!on)
	return(i_strdup("Did not find service."));

    mprintf(&string, "The services that depends on \"%s\":\n", on->name);

    while_active_db(current)
    {
        /* if current need the one we are watching */
        if (initng_depend(current, on) == TRUE)
        {
            if (current->current_state && current->current_state->state_name)
            {
                mprintf(&string, "  %s\t\t[%s]\n", current->name,
                        current->current_state->state_name);
            }
        }
    }
    return(string);
}

static char *cmd_print_fds(char *arg)
{
    char *string = NULL;
    active_db_h *currentA;
    s_call *currentC;
    process_h *currentP;
    int i;

    for (i = 0; i < 1024; i++)
    {

        currentC = NULL;
        while_list(currentC, &g.FDWATCHERS)
        {
            if (currentC->c.fdh->fds != i)
                continue;

            mprintf(&string, " %i: Used by plugin: %s\n", i, currentC->from_file);
            break;
            /* Call db fs */
        }

        currentA = NULL;
        while_active_db(currentA)
        {
            currentP = NULL;
            while_processes(currentP, currentA)
            {
                if (currentP->out_pipe[0] != i)
                    continue;

                mprintf(&string, " %i: Used service: %s, process: %s\n", i,
                        currentA->name, currentP->pt->name);
                break;
            }
        }


        /*mprintf(&string, " %i:\n", i); */
    }
    return(string);
}


static char *cmd_get_depends_off_deep(char *arg)
{
    char *string = NULL;
    active_db_h *current = NULL;
    active_db_h *on = NULL;

    on = initng_active_db_find_in_name(arg);

    if (!on)
	return(i_strdup("Did not find service."));

    mprintf(&string, "The the services that depends on \"%s\":\n", on->name);

    while_active_db(current)
    {
        /* if current need the one we are watching */
        if (initng_depend_deep(current, on) == TRUE)
        {
            if (current->current_state && current->current_state->state_name)
            {
                mprintf(&string, "  %s\t\t[%s]\n", current->name,
                        current->current_state->state_name);
            }
        }
    }
    return(string);
}

static int cmd_runlevel(char *arg)
{
    active_db_h *cur = NULL;

    /* check so option was provided */
    if (!arg)
        return (FALSE);

    /* search for the current runlevel by name */
    cur = initng_active_db_find_in_name(g.runlevel);


    /* start the new runlevel */
    if (!initng_handler_start_new_service_named(arg))
    {
        F_("New runlevel %s not found!\n", arg);
        return (FALSE);
    }

    /* stop old if found */
    if (cur)
        initng_handler_stop_service(cur);

    /* set new runlevel */
    initng_main_set_runlevel(arg);

    /* this function exists in another plugin and optional, will stop all services not in a runlevel */
    initng_command_execute('y');

    return (TRUE);
}

static int cmd_new_init(char *arg)
{
    char *new_i;
    int i = 0;

    if (!arg)
        return (FALSE);

    new_i = strdup(arg);
    g.new_init = split_n(new_i, &i, 0, 0);
    g.when_out = THEN_NEW_INIT;

    initng_handler_stop_all();
    return (TRUE);
}



#ifdef DEBUG


static int cmd_toggle_verbose(char *arg)
{
    (void) arg;
    switch (g.verbose)
    {
        case 0:
            g.verbose = 1;
            break;
        case 1:
            g.verbose = 0;
            break;
        case 2:
            g.verbose = 3;
            break;
        case 3:
            g.verbose = 2;
            break;
        default:
            g.verbose = 0;
            W_("Unknown verbose id %i\n", g.verbose);
            break;
    }
    return (g.verbose);
}

static int cmd_add_verbose(char *arg)
{
    if (!arg)
        return (FALSE);

    return (initng_error_verbose_add(arg));
}

static int cmd_del_verbose(char *arg)
{
    if (!arg)
        return (FALSE);

    return (initng_error_verbose_del(arg));
}
#endif

int module_init(int api_version)
{
    D_("module_init(stcmd);\n");
    if (api_version != API_VERSION)
    {
        F_("This module is compiled for api_version %i version and initng is compiled on %i version, won't load this module!\n", INITNG_VERSION, api_version);
        return (FALSE);
    }

    initng_command_add(&FAST_RELOAD);
    initng_command_add(&GET_PID_OF);
    initng_command_add(&START_ON_NEW);
    initng_command_add(&FREE_SERVICE);
    initng_command_add(&RESTART_SERVICE);
    initng_command_add(&PRINT_UPTIME);
    initng_command_add(&SERVICE_RELOAD);
    if (g.i_am_init)
    {
        initng_command_add(&REBOOT_INITNG);
        initng_command_add(&HALT_INITNG);
    }
    else
    {
        initng_command_add(&QUIT_INITNG);
    }
    initng_command_add(&PRINT_MODULES);
    initng_command_add(&LOAD_MODULE);
    initng_command_add(&UNLOAD_MODULE);
    initng_command_add(&PRINT_SERVICE_DB);
    initng_command_add(&PRINT_ACTIVE_DB);
    initng_command_add(&PERCENT_DONE);
    initng_command_add(&GET_FATHER_OF);
    initng_command_add(&DEPENDS_ON);
    initng_command_add(&DEPENDS_ON_DEEP);
    initng_command_add(&DEPENDS_OFF);
    initng_command_add(&DEPENDS_OFF_DEEP);
    initng_command_add(&NEW_INIT);
    initng_command_add(&CHANGE_RUNLEVEL);
    initng_command_add(&LIST_FDS);

#ifdef DEBUG
    initng_command_add(&TOGGLE_VERBOSE);
    initng_command_add(&ADD_VERBOSE);
    initng_command_add(&DEL_VERBOSE);
#endif


    D_("libstcmd.so.0.0 loaded!\n");
    return (TRUE);
}


void module_unload(void)
{
    D_("module_unload(stcmd);\n");

    initng_command_del(&FAST_RELOAD);
    initng_command_del(&GET_PID_OF);
    initng_command_del(&START_ON_NEW);
    initng_command_del(&FREE_SERVICE);
    initng_command_del(&RESTART_SERVICE);
    initng_command_del(&PRINT_UPTIME);
    initng_command_del(&SERVICE_RELOAD);
    if (g.i_am_init)
    {
        initng_command_del(&REBOOT_INITNG);
        initng_command_del(&HALT_INITNG);
    }
    else
    {
        initng_command_del(&QUIT_INITNG);
    }
    initng_command_del(&PRINT_MODULES);
    initng_command_del(&LOAD_MODULE);
    initng_command_del(&UNLOAD_MODULE);
    initng_command_del(&PRINT_SERVICE_DB);
    initng_command_del(&PRINT_ACTIVE_DB);
    initng_command_del(&PERCENT_DONE);
    initng_command_del(&GET_FATHER_OF);
    initng_command_del(&DEPENDS_ON);
    initng_command_del(&DEPENDS_ON_DEEP);
    initng_command_del(&DEPENDS_OFF);
    initng_command_del(&DEPENDS_OFF_DEEP);
    initng_command_del(&NEW_INIT);
    initng_command_del(&CHANGE_RUNLEVEL);
    initng_command_del(&LIST_FDS);
#ifdef DEBUG
    initng_command_del(&TOGGLE_VERBOSE);
    initng_command_del(&ADD_VERBOSE);
    initng_command_del(&DEL_VERBOSE);
#endif


    D_("libstcmd.so.0.0 unloaded!\n");

}
