/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>                          /* fcntl() */
#include "initng.h"
#include "initng_global.h"
#include "initng_service_cache.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"
#include "initng_signal.h"

#include "initng_plugin_callers.h"
#include "initng_plugin.h"

#include "initng_fd.h"

void initng_fd_close_all(void)
{
    s_call *current, *safe = NULL;

    S_;

    while_list_safe(current, &g.FDWATCHERS, safe)
    {
        close(current->c.fdh->fds);
        list_del(&current->list);

        if (current->from_file)
            free(current->from_file);
        free(current);
    }
}

/*
 * This function delivers what read to plugin,
 * through the hooks.
 * If no hook is found, or no return TRUE, it will
 * be printed to screen anyway.
 */
static void initng_fd_plugin_readpipe(active_db_h * service,
                                      process_h * process, char *buffer_pos)
{
    s_call *current = NULL;
    int delivered = FALSE;

    S_;


    while_list(current, &g.PIPEWATCHERS)
    {
        D_("Calling pipewatcher plugin.\n");
        if ((*current->c.pipewatcher) (service, process, buffer_pos) == TRUE)
            delivered = TRUE;
#ifdef DEBUG
        else
        {
            D_("plugin %s returned FALSE\n", current->from_file);
        }
#endif
    }

    /* make sure someone handled this */
    if (delivered != TRUE)
        fprintf(stdout, "%s", buffer_pos);
}


/*
 * This function is called when data is polled below,
 * or when a process is freed ( with flush_buffer set)
 */
void initng_fd_process_read_input(active_db_h * service, process_h * p)
{
    char *read_pos = NULL;
    int chars_read = 0;
    int old_content_offset = 0;
    int read_ret = 0;

    D_("\ninitng_fd_process_read_input(%s, %s, %i);\n", service->name,
       p->pt->name);

    if (p->out_pipe[0] <= 0)
    {
        F_("FIFO, can't be read! NOT OPEN!\n");
        return;
    }

    /* INITZIATE
     * if this are not set to out_pipe, if there is nothing to read. read() will block.
     * initng and sit down waiting for input.
     */
    if (p->buffer)
    {
        /* get the lenght of current data in buffer */
        old_content_offset = strlen(p->buffer);
    }
    else
    {
        /* initziate buffer fnctl */
        int fd_flags;

        fd_flags = fcntl(p->out_pipe[0], F_GETFL, 0);
        fcntl(p->out_pipe[0], F_SETFL, fd_flags | O_NONBLOCK);
    }

    /* read data from process, and continue again after a interrupt */
    do
    {
        errno = 0;

        /* OBSERVE, i_realloc may change the path to the data, so dont set buffer_pos to early */

        /* Make sure there is room for 100 more chars */
        D_("left: %i > %i\n", old_content_offset + chars_read + 100,
           p->buffer_allocated);
        if (old_content_offset + chars_read + 100 > p->buffer_allocated)
        {
            /* do a realloc */
            D_("Changing size of buffer %p to: %i\n", p->buffer,
               p->buffer_allocated + 100 + 1);
            char *tmp = i_realloc(p->buffer,
                                  (p->buffer_allocated + 100 +
                                   1) * sizeof(char));

            /* make sure realloc suceeded */
            if (tmp)
            {
                D_("p->buffer changes from %p to %p.\n", p->buffer, tmp);
                p->buffer = tmp;
                p->buffer_allocated += 100;
            }
            else
            {
                F_("realloc failed, possibly out of memory!\n");
                return;
            }
        }

        /* set read_pos to buffer + chars of old content + chars read so far */
        read_pos = p->buffer + old_content_offset + chars_read;

        /* read the data */
        D_("Reading 100 chars.\n");
        read_ret = read(p->out_pipe[0], read_pos, 100);

        D_("read_ret = %i\n", read_ret);
        /* make sure read does not return -1 */
        if (read_ret <= 0)
            break;

        /* make sure its nulled at end */
        read_pos[read_ret] = '\0';

        /* increase read_pos */
        chars_read += read_ret;
    }
    /* if read_ret == 100, it migit be more to read, or it got interrupted. */
    while (read_ret >= 100 || errno == EINTR);

    /* make sure there is any */
    if (chars_read > 0)
    {
        /* let all plugin take part of data */
        initng_fd_plugin_readpipe(service, p, p->buffer + old_content_offset);
    }

    /* if EOF close pipes */
    if (read_ret == 0)
    {
        D_("Closing fifos for %s.\n", service->name);
        if (p->out_pipe[0] > 0)
            close(p->out_pipe[0]);
        if (p->out_pipe[1] > 0)
            close(p->out_pipe[1]);
        p->out_pipe[0] = -1;
        p->out_pipe[1] = -1;
        return;
    }

    /* if buffer reached 10000 chars */
    if (old_content_offset + chars_read > 10000)
    {
        /* copy the string from 1000 chars, to first */
        memmove(p->buffer, &p->buffer[1000], 9000 * sizeof(char));
        /* rezise the buffer */
        p->buffer = i_realloc(p->buffer, 9000 * sizeof(char));
        p->buffer_allocated = 9000;
    }
}


/*
 * FILEDESCRIPTORPOLLNG
 *
 * File descriptors are scanned in 3 steps.
 * First, scan all plug-ins, and processes after file descriptors to
 * watch, and add them with FD_SET(); added++;
 *
 * Second, Run select if there where any added, with an timeout set
 * when function was called.
 *
 * Third, if select returns that there is data to fetch, walk over
 * plug-ins and file descriptors again, and call plug-ins or initng_process_readpipe
 * with pointer to plugin/service.
 *
 * This function will return FALSE, if it timeout, or if there was
 * no fds to monitor.
 */
int initng_fd_plugin_poll(int timeout)
{
    /* Variables */
    fd_set readset;
    struct timeval tv;
    int retval;
    int added = 0;
    active_db_h *currentA, *qA;
    s_call *currentC, *qC;
    process_h *currentP, *qP;

    /* initialization */
    S_;
    FD_ZERO(&readset);

    /* Set timeval struct, to our timeout */
    tv.tv_sec = timeout;
    tv.tv_usec = 0;


    /*
     * STEP 1:  Scan for fds to add
     */


    /* scan through active plug-ins that have listening file descriptors, and add them */
    currentC = NULL;
    while_list(currentC, &g.FDWATCHERS)
    {
        /* first, some checking */
        if (currentC->c.fdh->fds <= 2)
            continue;
        if (!currentC->c.fdh->call_module)
            continue;

        /*D_("adding fd #%i, from an call_db.\n", current->c.fdh->fds); */
        FD_SET(currentC->c.fdh->fds, &readset);
        added++;
    }

    /* Then add all file descriptors from process read pipes */
    currentA = NULL;
    while_active_db(currentA)
    {
        currentP = NULL;
        while_processes(currentP, currentA)
        {
            if (currentP->out_pipe[0] > 2)
            {
                FD_SET(currentP->out_pipe[0], &readset);
                added++;
            }
        }
    }


    /*
     * STEP 2: Do the select-poll, if any fds where added
     */

    /* check if there are any set */
    if (added <= 0)
    {
        D_("No file descriptors set.\n");
        sleep(timeout);
        return (TRUE);
    }
    D_("%i file descriptors added.\n", added);


    /* If interrupt then return and do main loop, it doesn't hurt if the file buffers have a bit more data anyway */
    if (g.interrupt)
        return (TRUE);

    /* make the select */
    retval = select(256, &readset, NULL, NULL, &tv);

    /* if interrupt is set, don't continue handling fussy fds */
    if (g.interrupt)
        return (TRUE);

    /* error */
    if (retval < 0)
    {
        F_("Select returned with %d %s\n", retval, strerror(errno));
        return (FALSE);
    }

    /* none was found */
    if (retval == 0)
    {
        D_("There was no data found on any added fd.\n");
        return (FALSE);
    }

    D_("%d fd's active\n", retval);

    /* If a fsck is running select will always return one file handler active, we give it 0.1 seconds to get some more data into the buffer. */
    sleep(0.1);


    /*
     * STEP 3:  Find the plug-ins/processes that handles the fd input, and call them
     */

    /* Now scan through callers */
    currentC = NULL;
    qC = NULL;
    while_list_safe(currentC, &g.FDWATCHERS, qC)
    {
        if (currentC->c.fdh->fds <= 2)
            continue;
        if (!currentC->c.fdh->call_module)
            continue;
        if (!FD_ISSET(currentC->c.fdh->fds, &readset))
            continue;

        D_("Calling plugin handler for fd #%i\n", currentC->c.fdh->fds);
        (*currentC->c.fdh->call_module) (currentC->c.fdh);
        D_("Call handler returned.\n");

        /* Found match, that means we need to look for one less, if we've found all we should then return */
        retval--;
        if (retval == 0)
            return (TRUE);
    }

    /* scan every service */
    currentA = NULL;
    qA = NULL;
    while_active_db_safe(currentA, qA)
    {
        currentP = NULL;
        qP = NULL;
        /* and all the processes */
        while_processes_safe(currentP, qP, currentA)
        {
            /* if matching */
            if (currentP->out_pipe[0] > 2
                && FD_ISSET(currentP->out_pipe[0], &readset))
            {
                D_("Will read from %s->start_process on fd #%i\n",
                   currentA->name, currentP->out_pipe[0]);
                /* Do the actual read from pipe */
                initng_fd_process_read_input(currentA, currentP);
                /* If we've been interrupted, return to main loop (program might have ended) */
                if (g.interrupt)
                    return (TRUE);
                /* Found match, that means we need to look for one less, if we've found all we should then return */
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
        }
    }

    return (TRUE);
}
